"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PgSortKeyCache = void 0;
const warp_contracts_1 = require("warp-contracts");
const pg_1 = require("pg");
class PgSortKeyCache {
    constructor(pgCacheOptions) {
        this.pgCacheOptions = pgCacheOptions;
        this.logger = warp_contracts_1.LoggerFactory.INST.create(PgSortKeyCache.name);
        this.setupPerformed = false;
        if (!pgCacheOptions.schemaName) {
            throw new Error('Schema name cannot be empty');
        }
        if (!pgCacheOptions.tableName) {
            throw new Error('Table name cannot be empty');
        }
        this.schemaName = pgCacheOptions.schemaName;
        this.tableName = pgCacheOptions.tableName;
        this.pool = new pg_1.Pool(pgCacheOptions).on('error', (err) => {
            this.logger.error('Unexpected error on idle client', err);
        });
    }
    async createTableIfNotExists() {
        await this.connection().query(`CREATE schema if not exists "${this.schemaName}"; SET search_path TO '${this.schemaName}';`);
        this.logger.info(`Attempting to create table ${this.tableName}`);
        const query = `
          CREATE TABLE IF NOT EXISTS "${this.tableName}"
          (
              id            bigserial,
              key           TEXT NOT NULL,
              sort_key      TEXT NOT NULL,
              value         JSONB,
              PRIMARY KEY (key, sort_key)
          );
          CREATE INDEX IF NOT EXISTS "idx_${this.tableName}_key_sk" ON "${this.tableName}" (key, sort_key DESC);
          CREATE INDEX IF NOT EXISTS "idx_${this.tableName}_key" ON "${this.tableName}" (key);`;
        await this.connection().query(query);
    }
    async begin() {
        this.logger.debug(`Begin transaction`);
        if (this.client == null) {
            this.client = await this.pool.connect();
        }
        await this.client.query('BEGIN;');
    }
    async close() {
        if (this.client) {
            this.client.release();
            this.client = null;
        }
        this.logger.info(`Connection released back to the pool`);
        return;
    }
    async cleanUp() {
        await this.close();
        await this.pool.end();
        this.pool = null;
        this.logger.info(`Pool cleaned up`);
    }
    async commit() {
        this.logger.debug(`Commit transaction`);
        if (this.client == null) {
            this.logger.error(`Called commit when no connection established.`);
            return;
        }
        await this.client.query('COMMIT;');
    }
    async delete(key) {
        await this.connection().query(`DELETE FROM "${this.schemaName}"."${this.tableName}" WHERE key = $1;`, [key]);
    }
    dump() {
        return Promise.resolve(undefined);
    }
    async get(cacheKey) {
        const result = await this.connection().query(`SELECT value
       FROM "${this.schemaName}"."${this.tableName}"
       WHERE key = $1
         AND sort_key = $2;`, [cacheKey.key, cacheKey.sortKey]);
        if (result && result.rows.length > 0) {
            return new warp_contracts_1.SortKeyCacheResult(cacheKey.sortKey, result.rows[0].value);
        }
        return null;
    }
    async getLast(key) {
        const result = await this.connection().query(`SELECT sort_key, value FROM "${this.schemaName}"."${this.tableName}" WHERE key = $1 ORDER BY sort_key DESC LIMIT 1;`, [key]);
        if (result && result.rows && result.rows.length > 0) {
            return new warp_contracts_1.SortKeyCacheResult(result.rows[0].sort_key, result.rows[0].value);
        }
        return null;
    }
    async getLastSortKey() {
        const result = await this.connection().query(`SELECT max(sort_key) as sort_key FROM "${this.schemaName}"."${this.tableName}";`);
        return result.rows[0].sort_key == '' ? null : result.rows[0].sortKey;
    }
    async getLessOrEqual(key, sortKey) {
        const result = await this.connection().query(`SELECT sort_key, value FROM "${this.schemaName}"."${this.tableName}" WHERE key = $1 AND sort_key <= $2 ORDER BY sort_key DESC LIMIT 1;`, [key, sortKey]);
        if (result && result.rows.length > 0) {
            return new warp_contracts_1.SortKeyCacheResult(result.rows[0].sort_key, result.rows[0].value);
        }
        return null;
    }
    async setUp() {
        const conf = this.pgCacheOptions;
        this.logger.info(`Connecting pg... ${conf.user}@${conf.host}:${conf.port}/${conf.database}`);
        await this.pool.query(`CREATE schema if not exists "${this.schemaName}"; SET search_path TO "${this.schemaName}";`);
        await this.createTableIfNotExists();
        this.logger.info(`Setup finished`);
    }
    async open() {
        if (!this.setupPerformed) {
            await this.setUp();
            this.setupPerformed = true;
        }
        if (!this.client) {
            this.client = await this.pool.connect();
        }
    }
    connection() {
        if (this.client) {
            return this.client;
        }
        return this.pool;
    }
    /**
     Let's assume that given contract cache contains these sortKeys: [a, b, c, d, e, f]
     Let's assume entriesStored = 2
     After pruning, the cache should be left with these keys: [e,f].
  
     const entries = await contractCache.keys({ reverse: true, limit: entriesStored }).all();
     This would return in this case entries [f, e] (notice the "reverse: true").
  
     await contractCache.clear({ lt: entries[entries.length - 1] });
     This effectively means: await contractCache.clear({ lt: e });
     -> hence the entries [a,b,c,d] are removed and left are the [e,f]
     */
    async prune(entriesStored = 5) {
        if (!entriesStored || entriesStored <= 0) {
            entriesStored = 1;
        }
        const allItems = +(await this.client.query(`SELECT count(1) AS total
         FROM "${this.schemaName}"."${this.tableName}"`)).rows[0].total;
        const deleted = +(await this.client.query(`
            WITH sorted_cache AS
                         (SELECT id, key, sort_key, row_number() over (PARTITION BY "key" ORDER BY sort_key DESC) AS rw
            FROM "${this.schemaName}"."${this.tableName}"), deleted AS
                (
            DELETE
            FROM "${this.schemaName}"."${this.tableName}"
            WHERE id IN (SELECT id FROM sorted_cache WHERE rw > $1) RETURNING *)
            SELECT count(1) AS del_total
            FROM deleted;
        `, [entriesStored])).rows[0].del_total;
        return {
            entriesBefore: allItems,
            entriesAfter: allItems - deleted,
            sizeBefore: -1,
            sizeAfter: -1
        };
    }
    async put(stateCacheKey, value) {
        const stringifiedValue = JSON.stringify(value);
        await this.removeOldestEntries(stateCacheKey.key);
        await this.connection().query(`
                INSERT INTO "${this.schemaName}"."${this.tableName}" (key, sort_key, value)
                VALUES ($1, $2, $3)
                ON CONFLICT(key, sort_key) DO UPDATE SET value = EXCLUDED.value`, [stateCacheKey.key, stateCacheKey.sortKey, stringifiedValue]);
    }
    async removeOldestEntries(key) {
        const rs = await this.connection().query(`
          SELECT count(1) as total
          FROM "${this.schemaName}"."${this.tableName}"
          WHERE key = $1
      `, [key]);
        if (rs.rows.length > 0) {
            const entriesTotal = rs.rows[0].total;
            if (entriesTotal >= this.pgCacheOptions.maxEntriesPerKey) {
                await this.connection().query(`
          WITH sorted_cache AS
                   (SELECT id, row_number() over (ORDER BY sort_key DESC) AS rw
                    FROM "${this.schemaName}"."${this.tableName}"
                    WHERE key = $1)
          DELETE
          FROM "${this.schemaName}"."${this.tableName}"
          WHERE id IN (SELECT id FROM sorted_cache WHERE rw >= $2);
      `, [key, this.pgCacheOptions.minEntriesPerKey]);
            }
        }
    }
    async rollback() {
        this.logger.debug(`Rollback`);
        if (this.client == null) {
            this.logger.error(`Rollback called, but no connection established`);
            return;
        }
        await this.client.query('ROLLBACK;');
    }
    storage() {
        return this.client;
    }
    async drop() {
        await this.client.query(`
          DROP INDEX IF EXISTS "idx_${this.tableName}_key_sk";
          DROP INDEX IF EXISTS "idx_${this.tableName}_key";
          DROP INDEX IF EXISTS "idx_${this.tableName}_owner";
          DROP TABLE IF EXISTS "${this.schemaName}"."${this.tableName}";
      `);
    }
    async batch(opStack) {
        try {
            await this.begin();
            for (const op of opStack) {
                if (op.type === 'put') {
                    await this.put(op.key, op.value);
                }
                else if (op.type === 'del') {
                    await this.delete(op.key);
                }
            }
            await this.commit();
        }
        catch (e) {
            await this.rollback();
            throw e;
        }
        finally {
            this.client.release();
            this.client = null;
        }
    }
    async del(cacheKey) {
        await this.connection().query(`
              INSERT INTO "${this.schemaName}"."${this.tableName}" (key, sort_key, value)
              VALUES ($1, $2, NULL)
              ON CONFLICT(key, sort_key) DO UPDATE SET value = EXCLUDED.value`, [cacheKey.key, cacheKey.sortKey]);
        return Promise.resolve(undefined);
    }
    async keys(sortKey, options) {
        const order = (options === null || options === void 0 ? void 0 : options.reverse) ? 'DESC' : 'ASC';
        const result = await this.connection().query({
            text: `WITH latest_values AS (SELECT DISTINCT ON (key) key, sort_key, value
                                     FROM "${this.schemaName}"."${this.tableName}"
                                     WHERE sort_key <= $1
                                       AND value IS NOT NULL
                                       AND ($2::text IS NULL OR key >= $2)
                                       AND ($3::text IS NULL OR key < $3)
                                     order by key ${order}, sort_key desc
                                     LIMIT $4::bigint)
              select key, value
              from latest_values
              order by key ${order};`,
            values: [sortKey, options === null || options === void 0 ? void 0 : options.gte, options === null || options === void 0 ? void 0 : options.lt, options === null || options === void 0 ? void 0 : options.limit],
            rowMode: 'array'
        });
        return result.rows.flat();
    }
    async kvMap(sortKey, options) {
        const order = (options === null || options === void 0 ? void 0 : options.reverse) ? 'DESC' : 'ASC';
        const result = await this.connection().query(`
              WITH latest_values AS (SELECT DISTINCT ON (key) key, sort_key, value
                                     FROM "${this.schemaName}"."${this.tableName}"
                                     WHERE sort_key <= $1
                                       AND value IS NOT NULL
                                       AND ($2::text IS NULL OR key >= $2)
                                       AND ($3::text IS NULL OR key < $3)
                                     order by key ${order}, sort_key desc
                                     LIMIT $4::bigint)
              select key, value
              from latest_values
              order by key ${order};`, [sortKey, options === null || options === void 0 ? void 0 : options.gte, options === null || options === void 0 ? void 0 : options.lt, options === null || options === void 0 ? void 0 : options.limit]);
        return new Map(result.rows.map((i) => [i.key, i.value]));
    }
}
exports.PgSortKeyCache = PgSortKeyCache;
//# sourceMappingURL=PgSortKeyCache.js.map